package base;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.lang.annotation.ElementType;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.Date;
import java.util.Iterator;
import java.util.TreeSet;

import base.CLI.CLIMenu;

public class CLI {

  private static String persistenceFile = "CLI2.ser";
  private Person currUser = null;
  private Collection<Network> networks = animate(persistenceFile);
  private IMenu menuState;
  
  /**
   * Should be able to:<ul>
   * <li>View Tasks ...</li>
   * <ul>
   * 	<li>... by chronological order.</li>
   * 	<li>... by deadline.</li>
   * </ul>
   * <li>Show repeating tasks once or all in range</li>
   * <li>Add new tasks.</li>
   * <li>Delete tasks.</li>
   * <li>Select tasks and ...</li>
   * <ul>
   * 	<li>... view their progress</li>
   * 	<li>... view their subtasks</li>
   * 	<li>... complete them either wholly or partially</li>
   * 	<li>... assign one or more users/networks to them</li>
   * </ul>
   * <li>View current networks.</li>
   * <li>Join/leave networks.</li>
   * <li>View tasks in network by claimed/unclaimed status and ...</li>
   * <ul>
   * 	<li>... by chronological order.</li>
   * 	<li>... by deadline.</li>
   * </ul>
   * @param args [not used]
   */
  public static void main(String[] args) {
  	new CLI();
  }
  
  /**
   * 
   * @author Paul Clark <pjc176 @ uclive.ac.nz></p>
   *
   * Annotation for use in creating CLI menus.</p>
   * 
   * Use this annotation with methods which should appear in the CLI menu. 
   * A menuText variable is required which is printed out to represent the 
   * annotated method in the menu. An optional weighting can be provided to 
   * force a particular option to a specific place in the list (the default 
   * weighting is 50). Usage is as follows:
   * <pre>@CLIMenu(menuText="Exit", weighting=99)
public void leave() {...}</pre>
	 * Or simply:<pre>@CLIMenu(menuText="Configure Profile")
public void config() {...}
   */
  @Retention(RetentionPolicy.RUNTIME)
  @Target(ElementType.METHOD)
  public @interface CLIMenu {
  	public String menuText();
  	public String menuID();
  	public int weighting() default 50;
  }
  
  public CLI() {
  	
  	menuState = new Menu("Main",this.getClass().getMethods());
  	
    if (networks == null) {
      networks = new ArrayList<Network>();
    }

    // Law of demeter
    if (!networks.isEmpty() && currUser==null) {
      currUser = networks.iterator().next().getMembers().iterator().next();
    }
    if (currUser == null) {
      currUser = signIn();
    } 

    System.out.printf("Hello %s!%n",currUser.getName());
    int swValue;
    for (;;) { // main loop, do this forever, or, until someone calls leave()
	    	try {
	      	menuState.parseArg().invoke(this, ((Object[]) null));
	    	} catch (java.lang.ArrayIndexOutOfBoundsException oob) {
	    		System.err.printf("Sorry, there is no valid menu entry corresponding to that index%n");
	    		//help();
	    		menuState.printMenu();
				} catch (IllegalArgumentException e) {
					System.err.println("Not enough arguments passed ...");
					e.printStackTrace();
				} catch (IllegalAccessException e) {
					e.printStackTrace();
				} catch (InvocationTargetException e) {
					e.printStackTrace();
				}
    } 
  } // end constructor
  
  /// methods for the menu(s) ///
  //TODO: Try moving the body of these methods to network and task, leave this class as an adapter?
  @CLIMenu(menuText="Add test data (users Jimmy and James)", menuID = "Main")
  public void addTestData() {
  	currUser = new Person("James");
  	Network nn = new Network("District");
  	Task nt = new Task();
  	nt.setName("Escape From mothership");
  	nt.setDescription("Damn I have been up here for way to long, I really have to escape");
  	nt.setSchedule("25/02/2011");
  	nt.setExpValue(9000);
  	nn.addMember(currUser);
  	nn.addTask(nt);
  	currUser.addTask(nt);
  	networks.add(nn);
  	
  	currUser = new Person("Jimmy");
  	currUser.addNetwork(nn);
  	nn = new Network("Flat");
  	nt = new Task();
  	nt.setName("Dishes");
  	nt.setDescription("Wash (most of) them before serving T/while cooking rice.");
  	nt.setSchedule("30/09/2010 1730");
  	nt.setExpValue(20);
  	nn.addMember(currUser);
  	nn.addTask(nt);
  	currUser.addTask(nt);
  	
  	nt = new Task();
  	nt.setName("Wake Up");
  	nt.setDescription("After sleeping, it is best to awaken.");
  	nt.setSchedule("01/10/2010 0730 +1d");
  	nt.setExpValue(20);
  	nn.addTask(nt);
  	currUser.addTask(nt);
  	
  	nt = new Task();
  	nt.setName("Some Old Task");
  	nt.setDescription("...");
  	nt.setSchedule("0730");
  	nt.setExpValue(2000);
  	nn.addTask(nt);
  	currUser.addTask(nt);

  	currUser.addNetwork(nn);
  	networks.add(nn);
  }

  @CLIMenu(menuText="Switch User", menuID = "Main")
  public void switchUser() {
  	//Switch User
    currUser = signIn();
  }
  
  //TODO:this method is too big, it smells
  @CLIMenu(menuText="Add a task", menuID = "Main Task")
  public Task addTask() {
    String strArg = "";
    int swValue = 1;
    Task newTask = new Task();
    strArg = Keyin.inString("Enter task name");
    newTask.setName(strArg);
    
    strArg = Keyin.inString("Enter description, be specific (hit enter to finish)");
    newTask.setDescription(strArg);
    
    System.out.println("Enter the schedule for the task");
    System.out.println("ie The following enters a task starting " +
                        "on the 25th of febuary 1989 at half past one in " +
                        "the afternoon, then repeating every three days");
                        System.out.println("\t25/02/1989 1330 +3d");
    strArg = Keyin.inString("Enter 'h' for more help");
    // API is too confusling
    if (strArg.matches("h")) {
      System.out.println("The schedule consist of three specifying " +
          "field, only one is required (either the time or date " +
          "of the task)\n their syntax is as follows:\n" +
          "\tdate (dd[/mm[/yyyy]])\n" +
          "\ttime (hh[:]mm) [ ((a.m.|am) | (p.m.|pm)) ]\n" +
          "\targ  + multiplier ((min,minutes)| (d|days) | " +
          "(w|weeks) | (m, months))\n" +
          "where '|' signifies an 'or', and square brakets '[...]' " +
          "enclose optional arguments\n" +
          "The order of the three fields can be as follows:\n" +
      "\t(time [date] | date [time]) arg*");
      strArg = Keyin.inString("Enter task schedule");
    }
    newTask.setSchedule(strArg);
    
    System.out.println("Enter how much this task is worth");
    swValue = Keyin.inInt("A rule of thumb is to make a task worth as " +
    "many points as it should take minutes to complete it");
    newTask.setExpValue(swValue);
    
    System.out.printf("Enter the network this this will be associated with%n" +
        "Press return for default (%s)%n",currUser.getDefaultNetwork().name());
    strArg = Keyin.inString();
    if (strArg.isEmpty()) {
      currUser.getDefaultNetwork().addTask(newTask);
    } else {
      // check the network exists and and user is a member of it
      for (Iterator<Network> iterator = networks.iterator(); iterator.hasNext();) {
        Network network = (Network) iterator.next();
        if (network.name().matches(strArg)) {
          if (network.contains(currUser.getName())) {
            network.addTask(newTask);
          } else {
            System.out.printf("The chosen network %s doesn't contain the user %s%n",network.name(),currUser.getName());
          }
          return null;
        }
        if (!iterator.hasNext()) {
          System.out.printf("Couldn't find a network called %s%n",strArg);
          return null;
        }
      }
    }
    currUser.addTask(newTask);
    // Testing
    System.out.println("New task created:");
    printTask(newTask);
    return newTask;
  }

  @CLIMenu(menuText="View your tasks", menuID = "Main Task")
  public void printTasks() {
    System.out.printf("%s's tasks:%n",currUser.getName());
    for (Iterator<Task> iterator = currUser.getToDoList().iterator(); iterator.hasNext();) {
      Task type = (Task) iterator.next();
      if (type.percentDone()!=100)
      	printTask(type);
      System.out.println("");
    }
    //menuState = changeState("Task");
    menuState = changeState(new TaskMenu(this.getClass().getMethods()));
  }
  
  @CLIMenu(menuText="Remove a task", menuID = "Task")
  public void removeTask() {
    // Remove tasks
    //viewTasks();
    //removeTask(taskName);
  }
  
  @CLIMenu(menuText="Execute a task", menuID = "Main Task")
  public void executeTask() {
  	System.out.println("Tasks on your ToDo list:");
  	Iterator<Task> todo = currUser.getToDoList().iterator();
  	while (todo.hasNext()) {
  		printTask(todo.next());
  	}
  	String taskName = Keyin.inString("Enter the name of the task to execute");
  	int points = currUser.executeTask(taskName);
  	if (points == -1) {
  		System.out.printf("Sorry the task %s isn't available to be completed%n",taskName);
  	} else {
  		System.out.printf("You got %d points!",points);
  	}
  }
  
  ///Networks///
  @CLIMenu(menuText="View your networks", menuID = "Main Network")
  public void viewNetworks() {
  	System.out.printf("%s's networks:%n",currUser.getName());
    for (Iterator<Network> iterator = currUser.getNetworksIt(); iterator.hasNext();) {
      Network type = (Network) iterator.next();
      System.out.printf("%s%n",type.name());
    }
    menuState = changeState("Network");
  }
  
  @CLIMenu(menuText="Join a network", menuID = "Network")
  public void joinNetwork() {
    System.out.println("Networks you can join");
    for (Iterator<Network> iterator = networks.iterator(); iterator.hasNext();) {
      Network type = (Network) iterator.next();
      if (type.getMember(currUser.getName())==null) {
        System.out.println(type.name());
      }
    }
    String netName = Keyin.inString("Enter the name of the network you wish to join");
    for (Iterator<Network> iterator = networks.iterator(); iterator.hasNext();) {
      Network type = (Network) iterator.next();
      if (type.name().toLowerCase().matches(netName.toLowerCase())) {
      	currUser.addNetwork(type);
      	System.out.printf("Welcome to %s!%n",type.name());
      	return;
      }
    }
    System.out.printf("Sorry, the network '%s' is not available to join.%n",netName);
  }
  
  @CLIMenu(menuText="Leave a network", menuID = "Network")
  public void leaveNetwork() {
  	System.out.println("Networks of which you are a member:");
  	Iterator<Network> netIt = currUser.getNetworksIt();
  	while (netIt.hasNext()) {
  		System.out.println(netIt.next().name());
  	}
  	String netName = Keyin.inString("Enter the name of the network you would like to leave");
  	currUser.removeNetworkByName(netName);
  }
  
  @CLIMenu(menuText="View all tasks within\n\t\tyour network(s)", menuID = "Network")
  public void viewNetworkTasks() {
    System.out.printf("Tasks within %s's networks:%n",currUser.getName());
    Iterator<Network> netIt = currUser.getNetworksIt();
    Network currNet = null;
    while (netIt.hasNext()) {
    	currNet = netIt.next();
    	System.out.printf("Network: %s%n",currNet.name());
    	Iterator<Task> taskIt = currNet.getTasks().iterator();
    	while (taskIt.hasNext()) {
    		printTask(taskIt.next());
    	}
    }
  }
  
  @CLIMenu(menuText="View unclaimed tasks\n\t\twithin your network(s)", menuID = "Task Network")
  public void viewUnclaimedNetworkTasks() {
  	System.out.printf("Tasks within %s's networks:%n",currUser.getName());
    Iterator<Network> netIt = currUser.getNetworksIt();
    Network currNet = null;
    while (netIt.hasNext()) {
    	currNet = netIt.next();
    	System.out.printf("Network: %s%n",currNet.name());
    	Iterator<Task> taskIt = currNet.getUnclaimedTasks().iterator();
    	while (taskIt.hasNext()) {
    		printTask(taskIt.next());
    	}
    }
  }

  @CLIMenu(menuText="Claim an available \n\t\tnetwork task", menuID = "Network")
  public void claimNetworkTasks() {
  	//TODO: implement this method
  	System.out.println("This method has not been implemented");
  }
  
  @CLIMenu(menuText="View users within your\n\t\tnetwork(s)", menuID = "Network")
  public void viewNetworkUsers() {
  	System.out.printf("Users within %s's networks:%n",currUser.getName());
    Iterator<Network> netIt = currUser.getNetworksIt();
    while (netIt.hasNext()) {
    	Iterator<Person> memberIt = netIt.next().getMembers().iterator();
    	while (memberIt.hasNext()) {
    		System.out.println(memberIt.next().getName());
    	}
    }
  }
  //end networks//
  
  @CLIMenu(menuText="Help", menuID = "Main Network Task", weighting=90)
  public void help() {
  	menuState.printMenu();
  }

  @CLIMenu(menuText="Exit", menuID = "Main", weighting=99)
  public void leave() {
    System.out.printf("Bye %s%n",currUser.getName());
    exit();
  }
  
  @CLIMenu(menuText="Back to main menu", menuID = "Network Task", weighting=99)
  public void goToMain() {
    menuState = new Menu("Main", this.getClass().getMethods());
  }
  
  /// Utility methods ///
  private Menu changeState(String newState) {
  	if (!newState.matches(((Menu) menuState).getState())) {
  		Keyin.inString("(Press ENTER to continue ...)");
  		return new Menu(newState, this.getClass().getMethods());
  	} else {
  		return (Menu) menuState;
  	}
  }
  
  //Re-implimenting state pattern to look more like GoF's one
  private IMenu changeState(IMenu newState) {
  		Keyin.inString("(Press ENTER to continue ...)");
  		return newState;
  }
  
  /**
   * Prints a list of all methods with a {@link CLIMenu} annotation, sorted 
   * according to the {@link MethodComparator}. The sorted list of methods is returned.</p>
   * 
   * <code>printMenu</code> is used to create an extensible CLI. Methods which are to be included 
   * in the CLI menu are annotated with the <code>CLIMenu</code> annotation. This 
   * annotation provides the text to be printed out to the menu for the entry 
   * representing the annotated method, and optionally a weighting.</br>
   * 
   * The weighting can be used to override the default (alphanumeric) ordering 
   * and force certain methods to the bottom or top of the list. 
   * ie,</br>
   * 	The 'Exit' option is traditionally the last item of a menu, to force this 
   * option to the bottom, it can be annotated as:
   * 		 <pre>@CLIMenu(menuText="Exit", weighting=99)
   * public void leave() {...}</pre>
   */
	private Method[] printMenu() {
  	System.out.println("MENU:");
  	Method[] meth = this.getClass().getMethods();
  	TreeSet<Method> methods = new TreeSet<Method>(new MethodComparator());
		CLIMenu ann = null;
  	for (int i = 0; i < meth.length; i++) {
			Method method = meth[i];
			ann = method.getAnnotation(CLIMenu.class);
			if (ann != null) {
				methods.add(method);
			}
		}
		int index = 0;

		// go through collection looking for annotation
		for (Iterator<Method> iterator = methods.iterator(); iterator.hasNext();) {
			Method method = (Method) iterator.next();
			ann = method.getAnnotation(CLIMenu.class);
				// then pretty print
				System.out.printf("\t%2d. %s%n",index++,ann.menuText());
		}
		return methods.toArray(new Method[methods.size()]);
  }

  /**
   * Change current system user.</p>
   *
   * Asks the user to input a user name. If a user by that name already exists 
   * then they become the current user, if no user by that name exists, the user 
   * is prompted to create a new account.
   *
   * @return The new current user
   */
  private Person signIn() {
    Person theUser = null;
    // get input from user
    String userName = newUserPrompt();
    // see if this user already exists
    theUser = getUser(userName);
    if (theUser==null) {
      System.out.printf("Couldn't find user %s%n",userName);
      String strArg = Keyin.inString("Would you like to create a new user? (Y|n)");
      if (strArg.isEmpty() || strArg.toLowerCase().matches("/^y.*/")) {
        theUser = newUser(userName);
      } else {
        exit();
      }
    }
    System.out.printf("You are now known as %s%n",userName);
    return theUser;
  }
  
  private String newUserPrompt() {
    // Display menu graphics
    System.out.println("==================================");
    System.out.println("|   Welcome ___________          |");
    System.out.println("==================================");
    System.out.println("| Please enter the name you will |");
    String strArg = Keyin.inString("|        be know by to continue  |");
    return strArg.isEmpty() ? "Anon" : strArg;
  }

  private Person newUser(String name) {
    Person newPerson = null;
    Network newNetwork = null;
    String strArg = name;

    if (strArg.isEmpty()) {
      newPerson = new Person("Anon");
      System.out.println("You are now known as \"Anon\"");
    } else {
      newPerson = new Person(strArg);
    }

    System.out.println("==================================");
    System.out.println("| Since this is your first time  |");
    System.out.println("| logging in please enter the    |");
    strArg = Keyin.inString("| name of your network           |");
    if (strArg.isEmpty()) {
      strArg = "Default";
    }
    for (Iterator<Network> iterator = networks.iterator(); iterator.hasNext();) {
      Network net = (Network) iterator.next();
      if (net.name().equals(strArg)) {
        newNetwork = net;
        break;
      }
    }
    if (newNetwork == null) {
      newNetwork = new Network(strArg);
    }
    newPerson.addNetwork(newNetwork);
    newPerson.setDefaultNetwork(newNetwork);
    newNetwork.addMember(newPerson);
    networks.add(newNetwork);

    return newPerson;
  }

  /**
   * Get existing user by name.
   *
   * @param  name The name of the user to be returned
   * @return The <code>Person</code> with the specified name, otherwise null.
   */
  private Person getUser(String name) {
    Person theUser = null;
    for (Iterator<Network> iterator = networks.iterator(); iterator.hasNext();) {
      Network network = (Network) iterator.next();
      theUser = network.getMember(name); 
      if (theUser != null) {
        break; // found them!
      }
    }
    return theUser;
  }

  private void printTask(Task type) {
    System.out.printf("\tName:\t%s%n",type.getName());
    System.out.printf("\tDescription:\t%s%n",type.getDescription());
    System.out.printf("\tNum. Increments\t%s%n",type.getNumPayouts());
    System.out.printf("\tExp. Value\t%s%n",type.getExpValue());
    System.out.printf("\tPercent Done\t%s%n",type.percentDone());
    System.out.printf("\tSchedule:%n");
    
    int TMI = 6;
    for(Date d : type.getSchedule()) {
    	if (TMI--==0) {
    		System.out.printf("\t\t...%n");
    		break;
    	}
      System.out.printf("\t\t%s%n",d.toString());
    }
  }

  private void exit() {
    serialize(persistenceFile, networks);
    System.exit(0);
  }

  /* Serialsiation methods adapted from here:
   * http://java.sun.com/developer/technicalArticles/Programming/serialization/ 
   */
  private static Collection<Network> animate(String fileName){
    Collection<Network> networks = new ArrayList<Network>();
    FileInputStream fis = null;
    ObjectInputStream in = null;
    try{
      fis = new FileInputStream(fileName);
      in = new ObjectInputStream(fis);
      networks = (Collection<Network>)in.readObject();
      in.close();
    }
    catch(IOException ex){
      ex.printStackTrace();
    }
    catch(ClassNotFoundException ex){
      ex.printStackTrace();
    }
    return networks;
  }

  private static void serialize(String fileName, Collection<Network> networks) {
    FileOutputStream fos = null;
    ObjectOutputStream out = null;
    try {
      fos = new FileOutputStream(fileName);
      out = new ObjectOutputStream(fos);
      out.writeObject(networks);
      out.close();
    }
    catch(IOException ex){
      ex.printStackTrace();
    }
  }
}

/**
 * @author      Paul Clark <pjc176 @ uclive.ac.nz>
 * @version     2010.1001
 * @since       1.6
 * 
 * @see #compare
 */
class MethodComparator implements Comparator<Method>, Serializable {
	
	/**
	 * 
	 */
	private static final long serialVersionUID = -3491536240580550154L;

	/**
   * Compare methods which have a valid {@link CLIMenu} annotation.
   *
   * Compares first on the weighting and then on the alphanumeric value of the menuText
   *
   * @param  m1 - the first method to be compared.
   * @param  m2 - the second method to be compared. 

   * @return a negative integer, zero, or a positive integer as the first 
   * argument is less than, equal to, or greater than the second. 
   */
	public int compare(Method m1, Method m2) {
		CLIMenu a1 = m1.getAnnotation(CLIMenu.class);
		CLIMenu a2 = m2.getAnnotation(CLIMenu.class);
		if( a1.weighting() > a2.weighting() ) {
			return 1;
		} else if( a1.weighting() < a2.weighting() ) {
			return -1;
		} else {
			return (m1.getAnnotation(CLIMenu.class).menuText().compareTo(m2.getAnnotation(CLIMenu.class).menuText()));
		}
	}
}

//**********************************************************
//**********************************************************
//Program: Keyin
//Reference: Session 20
//Topics:
// 1. Using the read() method of the ImputStream class
//    in the java.io package
// 2. Developing a class for performing basic console
//    input of character and numeric types
//**********************************************************
//**********************************************************

class Keyin {

  //*******************************
  //   support methods
  //*******************************
  //Method to display the user's prompt string
  public static void printPrompt(String prompt) {
    System.out.print(prompt + " ");
    System.out.flush();
  }

  //Method to make sure no data is available in the
  //input stream
  public static void inputFlush() {
    int dummy;
    int bAvail;

    try {
      while ((System.in.available()) != 0)
        dummy = System.in.read();
    } catch (java.io.IOException e) {
      System.out.println("Input error");
    }
  }

  //********************************
  //  data input methods for
  //string, int, char, and double
  //********************************
  public static String inString(String prompt) {
    inputFlush();
    printPrompt(prompt);
    return inString();
  }

  public static String inString() {
    int aChar;
    String s = "";
    boolean finished = false;

    while (!finished) {
      try {
        aChar = System.in.read();
        if (aChar < 0 || (char) aChar == '\n')
          finished = true;
        else if ((char) aChar != '\r')
          s = s + (char) aChar; // Enter into string
      }

      catch (java.io.IOException e) {
        System.out.println("Input error");
        finished = true;
      }
    }
    return s;
  }

  public static int inInt(String prompt) {
    while (true) {
      inputFlush();
      printPrompt(prompt);
      try {
        return Integer.valueOf(inString().trim()).intValue();
      }

      catch (NumberFormatException e) {
        System.out.println("Invalid input. Not an integer");
      }
    }
  }

  public static char inChar(String prompt) {
    int aChar = 0;

    inputFlush();
    printPrompt(prompt);

    try {
      aChar = System.in.read();
    }

    catch (java.io.IOException e) {
      System.out.println("Input error");
    }
    inputFlush();
    return (char) aChar;
  }

  public static double inDouble(String prompt) {
    while (true) {
      inputFlush();
      printPrompt(prompt);
      try {
        return Double.valueOf(inString().trim()).doubleValue();
      }

      catch (NumberFormatException e) {
        System.out
        .println("Invalid input. Not a floating point number");
      }
    }
  }
}
