package base;

	/*
	 * ==================================================================== The
	 * Apache Software License, Version 1.1
	 *
	 * Copyright (c) 2003 Digital Clash LLC. All rights reserved.
	 *
	 * Redistribution and use in source and binary forms, with or without
	 * modification, are permitted provided that the following conditions are met: 1.
	 * Redistributions of source code must retain the above copyright notice, this
	 * list of conditions and the following disclaimer. 2. Redistributions in
	 * binary form must reproduce the above copyright notice, this list of
	 * conditions and the following disclaimer in the documentation and/or other
	 * materials provided with the distribution. 3. The end-user documentation
	 * included with the redistribution, if any, must include the following
	 * acknowledgment: "This product includes software developed by the ChronicJ
	 * team (http://www.chronicj.org/)." Alternately, this acknowledgment may
	 * appear in the software itself, if and wherever such third-party
	 * acknowledgments normally appear. 4. The names "ChronicJ" and "Digital Clash"
	 * not be used to endorse or promote products derived from this software
	 * without prior written permission. For written permission, please contact
	 * info@digitalclash.com. 5. Products derived from this software may not be
	 * called "ChronicJ", "Digital Clash", nor may "ChronicJ" or "Digital Clash"
	 * appear in their name, without prior written permission of Digital Clash LLC.
	 *
	 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
	 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
	 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
	 * APACHE SOFTWARE FOUNDATION OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
	 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
	 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
	 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
	 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
	 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
	 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
	 * ==================================================================== This
	 * product includes software developed by the by the Apache Software Foundation
	 * (http://www.apache.org/).
	 * ====================================================================
	 */

	import org.chronicj.DateRange;
	import org.chronicj.Event;
	import org.chronicj.RangeIterator;
	import org.chronicj.Schedule;
	import org.chronicj.ScheduleElement;
	import org.chronicj.TemporalExpression;
	import org.chronicj.TimePoint;
import org.chronicj.impl.DefaultScheduleElement;

	import java.io.Serializable;
	import java.util.ArrayList;
	import java.util.HashMap;
	import java.util.Iterator;
	import java.util.List;
import java.util.Map;


	/**
	 * Default implementation of a {@link org.chronicj.Schedule Schedule}.
	 *
	 * @author <a href="mlipper@US-ABP.com">Matthew Lipper</a>
	 *
	 * @see <a href="http://martinfowler.com/apsupp/recurring.pdf">Recurring
	 *      Events for Calendars</a>
	 */
	public class FullSchedule implements Schedule, Serializable {
	    private Map elements;
	    static final long serialVersionUID=42L;
	    /**
	     * Creates an instance of a <code>FullSchedule</code>.
	     */
	    public FullSchedule() {
	        super();
	        elements = new HashMap();
	    }

	    /**
	     * Default implementation of
	     * {@link org.chronicj.Schedule#isOccuring(org.chronicj.Event,org.chronicj.DateRangeImpl)}.
	     *
	     * @param anEvent
	     *            the Event whose schedule is being queried
	     * @param aDateRange
	     *            the time period in which to check schedule information
	     *
	     * @return a boolean value indicating whether the event occurs within the
	     *         specified DateRange
	     *
	     * @throws NullPointerException
	     *             if either argument is null
	     * @throws IllegalArgumentException
	     *             if the DateRange argument is empty
	     *
	     * @see org.chronicj.Schedule#isOccuring(org.chronicj.Event,org.chronicj.DateRangeImpl)
	     */
	    public boolean isOccuring(Event anEvent, DateRange aDateRange) {
	        if (!isScheduled(anEvent)) {
	            return false;
	        }

	        for (Iterator i = elements.values().iterator(); i.hasNext();) {
	            for (Iterator j = ((List) i.next()).iterator(); j.hasNext();) {
	                if (((ScheduleElement) j.next()).isOccuring(anEvent, aDateRange)) {
	                    return true;
	                }
	            }
	        }

	        return false;
	    }

	    /**
	     * Default implementation of
	     * {@link org.chronicj.Schedule#isOccuring(org.chronicj.Event,org.chronicj.TimePoint)}.
	     *
	     * @param anEvent
	     *            the Event whose schedule is being queried
	     * @param aTimePoint
	     *            the moment in time
	     *
	     * @return a boolean value indicating whether the event occurs during the
	     *         specified TimePoint
	     *
	     * @throws NullPointerException
	     *             if either argument is null
	     *
	     * @see org.chronicj.Schedule#isOccuring(org.chronicj.Event,org.chronicj.DateRangeImpl)
	     */
	    public boolean isOccuring(Event anEvent, TimePoint aTimePoint) {
	        if (!isScheduled(anEvent)) {
	            return false;
	        }

	        for (Iterator i = elements.values().iterator(); i.hasNext();) {
	            for (Iterator j = ((List) i.next()).iterator(); j.hasNext();) {
	                if (((ScheduleElement) j.next()).isOccuring(anEvent, aTimePoint)) {
	                    return true;
	                }
	            }
	        }

	        return false;
	    }

	    public boolean isScheduled(Event anEvent) {
	        return elements.containsKey(anEvent.getId());
	    }

	    public void add(Event anEvent, TemporalExpression aTemporalExpression) {
	        List scheduleElements = null;

	        if (!isScheduled(anEvent)) {
	            scheduleElements = new ArrayList();
	            scheduleElements.add(new DefaultScheduleElement(anEvent,
	                    aTemporalExpression));
	            elements.put(anEvent.getId(), scheduleElements);
	        } else {
	            scheduleElements = (List) elements.get(anEvent.getId());

	            ScheduleElement newScheduleElement = new DefaultScheduleElement(anEvent,
	                    aTemporalExpression);

	            if (!scheduleElements.contains(newScheduleElement)) {
	                scheduleElements.add(newScheduleElement);
	            }
	        }
	    }

	    /**
	     * All calculations are made using {@link DatePrecision#DATE} precision.
	     *
	     * @see org.chronicj.Schedule#dates(org.chronicj.Event,
	     *      org.chronicj.DateRangeImpl)
	     */
	    public List dates(Event anEvent, DateRange during) {
	        if (!isScheduled(anEvent)) {
	            return null;
	        }

	        List results = new ArrayList();

	        for (RangeIterator rangeIter = new RangeIterator(during);
	                rangeIter.hasNext();) {
	            DateRange subRange = (DateRange) rangeIter.next();
	            List anEventSchedule = (List) elements.get(anEvent.getId());

	            for (Iterator scheduleIter = anEventSchedule.iterator();
	                    scheduleIter.hasNext();) {
	                ScheduleElement element = (ScheduleElement) scheduleIter.next();
	                System.out.println("Checking schedule element " + element +
	                    " against range " + subRange);

	                if (element.isOccuring(anEvent, subRange.start())) {
	                    results.add(subRange.start());
	                }
	            }
	        }

	        return results;
	    }

	    /*
	     * @see org.chronicj.Schedule#nextOccurence(org.chronicj.Event,
	     *      org.chronicj.TimePoint)
	     */
	    public TimePoint nextOccurence(Event anEvent, TimePoint aTimePoint) {
	        // TODO Auto-generated method stub
	        return null;
	    }

}
