package org.chronicj;

import java.io.ObjectStreamException;
import java.io.Serializable;

import java.util.Arrays;
import java.util.Calendar;
import java.util.Collections;
import java.util.List;

/**
 * Constants used for expressing date precision when working with ChronicJ
 * components. Implemented using the <a
 * href="http://developer.java.sun.com/developer/Books/shiftintojava/page1.html">
 * Typesafe Enum pattern</a> popularized by Joshua Bloch.
 * 
 * @author <a href="mlipper@US-ABP.com">Matthew Lipper</a>
 */
public abstract class DatePrecision implements Comparable, Serializable
{
	public static final DatePrecision YEAR = new DatePrecision(Calendar.YEAR)
	{
		TimePoint increment(TimePoint aTimePoint, int amount)
		{
			return aTimePoint.addYears(amount);
		}
	};

	public static final DatePrecision MONTH = new DatePrecision(Calendar.MONTH)
	{
		TimePoint increment(TimePoint aTimePoint, int amount)
		{
			return aTimePoint.addMonths(amount);
		}
	};

	public static final DatePrecision DATE = new DatePrecision(Calendar.DATE)
	{
		TimePoint increment(TimePoint aTimePoint, int amount)
		{
			return aTimePoint.addDays(amount);
		}
	};

	public static final DatePrecision HOUR_OF_DAY =
		new DatePrecision(Calendar.HOUR_OF_DAY)
	{
		TimePoint increment(TimePoint aTimePoint, int amount)
		{
			return aTimePoint.addHours(amount);
		}
	};

	public static final DatePrecision MINUTE =
		new DatePrecision(Calendar.MINUTE)
	{
		TimePoint increment(TimePoint aTimePoint, int amount)
		{
			return aTimePoint.addMinutes(amount);
		}
	};

	public static final DatePrecision SECOND =
		new DatePrecision(Calendar.SECOND)
	{
		TimePoint increment(TimePoint aTimePoint, int amount)
		{
			return aTimePoint.addSeconds(amount);
		}
	};

	public static final DatePrecision MILLISECOND =
		new DatePrecision(Calendar.MILLISECOND)
	{
		TimePoint increment(TimePoint aTimePoint, int amount)
		{
			return aTimePoint.addMilliSeconds(amount);
		}
	};

	private static final DatePrecision[] PRECISION =
		{ YEAR, MONTH, DATE, HOUR_OF_DAY, MINUTE, SECOND, MILLISECOND };
	public static final List VALUES =
		Collections.unmodifiableList(Arrays.asList(PRECISION));
	private final int precision;

	private DatePrecision(int calendarField)
	{
		precision = calendarField;
	}

	public int compareTo(Object arg)
	{
		if (arg == null)
		{
			throw new NullPointerException("Object argument cannot be null.");
		}

		//Unchecked cast: runtime error is desired if not proper type
		DatePrecision other = (DatePrecision) arg;

		//Sanity check (consistent w/ equals)
		if (this.equals(other) && (this.precision == other.precision))
		{
			return 0;
		}

		if (this.precision > other.precision)
		{
			return 1;
		}

		return -1;
	}

	public boolean greaterThan(DatePrecision another)
	{
		return compareTo(another) == 1;
	}

	public boolean lessThan(DatePrecision another)
	{
		return compareTo(another) == -1;
	}

	abstract TimePoint increment(TimePoint aTimePoint, int amount);

	public String toString()
	{
		switch (precision)
		{
			case Calendar.YEAR :
				return "YEAR";

			case Calendar.MONTH :
				return "MONTH";

			case Calendar.DATE :
				return "DATE";

			case Calendar.HOUR_OF_DAY :
				return "HOUR_OF_DAY";

			case Calendar.MINUTE :
				return "MINUTE";

			case Calendar.SECOND :
				return "SECOND";

			case Calendar.MILLISECOND :
				return "MILLISECOND";

			default :
				throw new IllegalStateException("DatePrecision unknown. Programmer error!");
		}
	}

	private Object readResolve() throws ObjectStreamException
	{
		return PRECISION[precision];
	}
}
