/*
 * ==================================================================== The
 * Apache Software License, Version 1.1
 *
 * Copyright (c) 2003 Digital Clash LLC. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 1.
 * Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer. 2. Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution. 3. The end-user documentation
 * included with the redistribution, if any, must include the following
 * acknowledgment: "This product includes software developed by the ChronicJ
 * team (http://www.chronicj.org/)." Alternately, this acknowledgment may
 * appear in the software itself, if and wherever such third-party
 * acknowledgments normally appear. 4. The names "ChronicJ" and "Digital Clash"
 * not be used to endorse or promote products derived from this software
 * without prior written permission. For written permission, please contact
 * info@digitalclash.com. 5. Products derived from this software may not be
 * called "ChronicJ", "Digital Clash", nor may "ChronicJ" or "Digital Clash"
 * appear in their name, without prior written permission of Digital Clash LLC.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * APACHE SOFTWARE FOUNDATION OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ==================================================================== This
 * product includes software developed by the by the Apache Software Foundation
 * (http://www.apache.org/).
 * ====================================================================
 */
package org.chronicj;

import java.util.Arrays;
import java.util.Date;


/**
 * Defines basic requirements and utility methods for classes that model
 * periods of time. Based directly on the <a
 * href="http://martinfowler.com/ap2/range.html">Range</a> pattern described
 * by Martin Fowler. This is a value object.
 *
 * @author <a href="mlipper@US-ABP.com">Matthew Lipper</a>
 *
 * @see org.chronicj.TimePoint
 * @see <a href="http://martinfowler.com/ap2/range.html">Range</a>
 */
public class DateRange implements Comparable
{
    /** Specifies an empty range which can be used as a constant */
    public static final DateRange EMTPY = new DateRange(new TimePoint(1969, 5, 8),
            new TimePoint(1968, 12, 12));
    private TimePoint end;
    private TimePoint start;

    /**
     * Construct a DateRange using the given start and end dates. The {@link
     * java.util.Date} arguments are wrapped using {@link TimePoint}s, and as
     * such, they are normalized using minute precision (this is the default
     * precision for a TimePoint). If the start date is greater than the end
     * date, the range is considered empty. Currently, this class does not
     * support open ranges, i.e. since or until.
     *
     * @param aStartDate
     *                    the Date that is the lower end of the range
     * @param anEndDate
     *                    the Date that is the upper end of the range
     *
     * @throws NullPointerException
     *                    is either argument is null
     */
    public DateRange(Date aStartDate, Date anEndDate)
    {
        this((aStartDate != null) ? new TimePoint(aStartDate) : null,
            (anEndDate != null) ? new TimePoint(anEndDate) : null);
    }

    /**
     * Construct a DateRange using the supplied start and end <code>TimePoint</code>s.
     * If the start date is greater than the end date, the range is considered
     * empty. Currently, this class does not support open ranges, i.e. since or
     * until.
     *
     * @param aStartDate
     *                    the TimePoint that is the lower end of the range
     * @param anEndDate
     *                    the TimePoint that is the upper end of the range
     *
     * @throws NullPointerException
     *                    is either argument is null
     */
    public DateRange(TimePoint aStartDate, TimePoint anEndDate)
    {
        if (aStartDate == null)
        {
            throw new NullPointerException("Lower range cannot be null.");
        }

        if (anEndDate == null)
        {
            throw new NullPointerException("Upper range cannot be null.");
        }

        start = aStartDate;
        end = anEndDate;
    }

    public static boolean isContiguous(DateRange[] args)
    {
        Arrays.sort(args);

        for (int i = 0; i < (args.length - 1); i++)
        {
            if (!args[i].abuts(args[i + 1]))
            {
                return false;
            }
        }

        return true;
    }

    /**
     * The {@link DatePrecision}of this range. If the start and end of this
     * range are of different precisions, the greater precision is returned.
     *
     * @return DatePrecision of this range
     */
    public DatePrecision getDatePrecision()
    {
        if (start.getDatePrecision().greaterThan(end.getDatePrecision()))
        {
            return start.getDatePrecision();
        }

        //Less than or equal
        return end.getDatePrecision();
    }

    /**
     * Used mostly by other date range calculations to indicate an empty set.
     *
     * @return true or false, indicating whether the start date occurs after
     *               the end date
     */
    public boolean isEmpty()
    {
        return start.after(end);
    }

    /**
     * Used to detect whether two date ranges abut each other.
     *
     * @param anotherRange
     *                    the range to check against
     *
     * @return true if the supplied DateRange argument abuts this range either
     *               occurring before or after
     */
    public boolean abuts(DateRange anotherRange)
    {
        return !overlaps(anotherRange) && gap(anotherRange).isEmpty();
    }

    public static DateRange combination(DateRange[] args)
    {
        Arrays.sort(args);

        if (!isContiguous(args))
        {
            throw new IllegalArgumentException("Unable to combine date ranges");
        }

        return new DateRange(args[0].start(), args[args.length - 1].end());
    }

    /**
     * Compare the current DateRange instance to the one provided. The
     * following algorithms are used if the start dates are not equal:
     *
     * <ul>
     * <li>whichever instance has a later start date (as determined by {@link
     * #start()} is the greater value</li>
     * </ul>
     *
     * if the start dates are equal then:
     *
     * <ul>
     * <li>whichever instance has a later end date (as determined by {@link
     * #end()} is the greater value</li>
     * <li>if the end dates are equal, then the instances are equal</li>
     * </ul>
     *
     * <b>NOTE:</b> results are non-deterministic if either DateRange
     * isEmpty()
     *
     * @param arg
     *                    Object to compare
     *
     * @return <ul>
     *               <li>-1 if this TimePoint occurs after the one supplied</li>
     *               <li>0 if this TimePoint occurs at the same time as the one
     *               provided</li>
     *               <li>1 if this TimePoint occurs after the one provided</li>
     *               </ul>
     *
     * @throws ClassCastException
     *                    if argument is not a DateRange
     * @throws NullPointerException
     *                    if argument is null
     *
     * @see {@link Comparable#compareTo(java.lang.Object)}
     */
    public int compareTo(Object arg)
    {
        DateRange other = (DateRange) arg;

        if (!start.equals(other.start()))
        {
            return start.compareTo(other.start());
        }

        return end.compareTo(other.end());
    }

    /**
     * Accessor method for the <code>TimePoint</code> denoting the upper end
     * of this range.
     *
     * @return the TimePoint which holds the current value of the end field
     */
    public TimePoint end()
    {
        return end;
    }

    /*
     * True if the supplied argument ia an instance of DateRange whose start()
     * method returns a value equal this.start and whose end() method returns a
     * value equal to this.end.
     *
     * @see java.lang.Object#equals(java.lang.Object)
     */
    public boolean equals(Object arg)
    {
        if (!(arg instanceof DateRange))
        {
            return false;
        }

        DateRange other = (DateRange) arg;

        return start.equals(other.start()) && end.equals(other.end());
    }

    /**
     * Used to determine the duration of time (expressed as another DateRange)
     * between the current DateRange instance and another DateRange. For now,
     * the default precision, {@link DatePrecision#MINUTE}, is used.
     *
     * @param arg the other DateRange against which to determine an intervening gap
     *
     * @return anotherRange DateRange delimiting beginning and ending times of
     *               the gap
     */
    public DateRange gap(DateRange arg)
    {
        if (this.overlaps(arg))
        {
            return DateRange.EMTPY;
        }

        DateRange lower;
        DateRange higher;

        if (this.compareTo(arg) < 0)
        {
            lower = this;
            higher = arg;
        }
        else
        {
            lower = arg;
            higher = this;
        }

        return new DateRange(lower.end().addMinutes(1),
            higher.start().addMinutes(-1));
    }

    /*
     * Uses hashCode() of the start field.
     *
     * @see java.lang.Object#hashCode()
     */
    public int hashCode()
    {
        return start.hashCode();
    }

    /**
     * Used to check whether a given <code>TimePoint</code> occurs within the
     * current range.
     *
     * @param aDate
     *                    the TimePoint that is to be checked for occurence within the
     *                    current range
     *
     * @return true or false, indicating whether the supplied Date occurs
     *               within the current range
     */
    public boolean includes(TimePoint aDate)
    {
        return !aDate.before(start) && !aDate.after(end);
    }

    /**
     * Used to check whether a given <code>DateRange</code> occurs within the
     * current range.
     *
     * @param arg
     *                    the DateRange that is to be checked for occurence within the
     *                    current range
     *
     * @return true or false, indicating whether the supplied DateRange occurs
     *               within the current range
     */
    public boolean includes(DateRange arg)
    {
        return this.includes(arg.start()) && this.includes(arg.end());
    }

    /**
     * Used mostly by other date range calculations to indicate an empty set.
     *
     * @param anotherRange
     *                    the other DateRange against which to determine an intervening
     *                    gap
     *
     * @return true or false, indicating whether the start date occurs after
     *               the end date
     */
    public boolean overlaps(DateRange anotherRange)
    {
        return anotherRange.includes(start) || anotherRange.includes(end) ||
        includes(anotherRange);
    }

    /**
     * Used to check whether a group of ranges completely partition the current
     * DateRange. For this to be true, the given set of DateRange
     * arguments,must be contiguous with respect to each other and, when
     * combined, equal the current range.
     *
     * @param otherRanges
     *                    the set of DateRange objects to check against
     *
     * @return true if the supplied ranges completely partition <code>this</code>
     *               range, false otherwise
     */
    public boolean partitionedBy(DateRange[] otherRanges)
    {
        if (!DateRange.isContiguous(otherRanges))
        {
            return false;
        }

        return this.equals(DateRange.combination(otherRanges));
    }

    /**
     * Accessor method for the <code>TimePoint</code> denoting the lower end
     * of this range.
     *
     * @return the TimePoint which holds the current value of the start field
     */
    public TimePoint start()
    {
        return start;
    }

    /*
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
        if (isEmpty())
        {
            return "Empty Date Range";
        }

        return start.toString() + " - " + end.toString();
    }
}
