/*
 * ==================================================================== The
 * Apache Software License, Version 1.1
 * 
 * Copyright (c) 2003 Digital Clash LLC. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 1.
 * Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer. 2. Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution. 3. The end-user documentation
 * included with the redistribution, if any, must include the following
 * acknowledgment: "This product includes software developed by the ChronicJ
 * team (http://www.chronicj.org/)." Alternately, this acknowledgment may
 * appear in the software itself, if and wherever such third-party
 * acknowledgments normally appear. 4. The names "ChronicJ" and "Digital Clash"
 * not be used to endorse or promote products derived from this software
 * without prior written permission. For written permission, please contact
 * info@digitalclash.com. 5. Products derived from this software may not be
 * called "ChronicJ", "Digital Clash", nor may "ChronicJ" or "Digital Clash"
 * appear in their name, without prior written permission of Digital Clash LLC.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * APACHE SOFTWARE FOUNDATION OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ==================================================================== This
 * product includes software developed by the by the Apache Software Foundation
 * (http://www.apache.org/).
 * ====================================================================
 */
package org.chronicj;

import java.util.Iterator;

/**
 * <p>
 * Simple {@link java.util.Iterator Iterator}implmentation that can be used to
 * traverse the sub ranges that occur between the start and end
 * {@link org.chronicj.TimePoint TimePoints}of a given
 * {@link org.chronicj.DateRange DateRange}.
 * </p>
 * <p>
 * Iteration proceeds using the following semantics: <br />
 * <ul>
 * <li>{@link hasNext()}returns true if the <code>DateRange</code> returned
 * by calling {@link next()}is included by the current range.</li>
 * <li>{@link next()}returns a new <code>DateRange</code> instance created
 * by using a start date of <code>current start date +1</code> and an end
 * date of <code>current start date +2</code>.<b>NOTE:</b> increment will
 * reflect the {@link org.chronicj.DatePrecision DatePrecision}of the
 * underlying <code>TimePoint</code> s with which the current range was
 * created.</li>
 * <li>{@link remove()}is not supported.</li>
 * </ul>
 * </p>
 * 
 * @author <a href="mlipper@US-ABP.com">Matthew Lipper</a>
 * 
 * @see org.chronicj.DateRange
 * @see org.chronicj.TimePoint
 * @see java.util.Iterator
 * @see <a href="http://martinfowler.com/ap2/range.html">Range</a>
 */
public class RangeIterator implements Iterator
{
	private final DateRange range;
	private DateRange next;

	public RangeIterator(DateRange aDateRange)
	{
		range = aDateRange;
	}

	/**
	 * Returns: <br/><code>currentRange.includes(nextRangeIndex));</code>
	 * 
	 * @see java.util.Iterator#hasNext()
	 */
	public boolean hasNext()
	{
		if (range.isEmpty())
		{
			return false;
		}

		if (next == null)
		{
			return range.includes(
				new DateRange(
					range.start().increment(1),
					range.start().increment(2)));
		}

		return range.includes(
			new DateRange(
				next.start().increment(1),
				next.start().increment(2)));
	}

	/**
	 * Returns: <br/><code>new DateRange(currentRangeIndex.start().increment(1), currentRangeIndex.start().increment(2));</code>
	 * 
	 * @see java.util.Iterator#next()
	 */
	public Object next()
	{
		if (next == null)
		{
			//First
			next = new DateRange(range.start(), range.start().increment(1));

			return next;
		}

		next =
			new DateRange(next.start().increment(1), next.start().increment(2));

		return next;
	}

	/**
	 * Currently, throws an <code>UnsupportedOperationException</code> as
	 * this method is not supported.
	 * 
	 * @see java.util.Iterator#remove()
	 */
	public void remove()
	{
		throw new UnsupportedOperationException();
	}
}
