/*
 * ==================================================================== The
 * Apache Software License, Version 1.1
 * 
 * Copyright (c) 2003 Digital Clash LLC. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 1.
 * Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer. 2. Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution. 3. The end-user documentation
 * included with the redistribution, if any, must include the following
 * acknowledgment: "This product includes software developed by the ChronicJ
 * team (http://www.chronicj.org/)." Alternately, this acknowledgment may
 * appear in the software itself, if and wherever such third-party
 * acknowledgments normally appear. 4. The names "ChronicJ" and "Digital Clash"
 * not be used to endorse or promote products derived from this software
 * without prior written permission. For written permission, please contact
 * info@digitalclash.com. 5. Products derived from this software may not be
 * called "ChronicJ", "Digital Clash", nor may "ChronicJ" or "Digital Clash"
 * appear in their name, without prior written permission of Digital Clash LLC.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * APACHE SOFTWARE FOUNDATION OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ==================================================================== This
 * product includes software developed by the by the Apache Software Foundation
 * (http://www.apache.org/).
 * ====================================================================
 */
package org.chronicj;

import java.util.List;

/**
 * <p>
 * Represents a coherent collection of logical events that are expected to
 * occur over the course of time. A <code>Schedule</code> is composed of zero
 * or more {@link org.chronicj.ScheduleElement ScheduleElement}s which are
 * queried to handle requests for schedule information.
 * </p>
 * 
 * <p>
 * Schedule also provides the public API for adding, editing and removing
 * scheduled events for a given entity.
 * </p>
 * 
 * <p>
 * This class is based directly on patterns described in a paper by Martin
 * Fowler which can be found <a
 * href="http://martinfowler.com/apsupp/recurring.pdf">here</a>.
 * </p>
 * 
 * @author <a href="mlipper@US-ABP.com">Matthew Lipper</a>
 * 
 * @see <a href="http://martinfowler.com/apsupp/recurring.pdf">Recurring
 *           Events for Calendars</a>
 */
public interface Schedule
{
	/**
	 * Determines if a given event occurs during a particular instant in time.
	 * 
	 * @param anEvent
	 *                    the Event whose schedule is being queried
	 * @param aTimePoint
	 *                    the moment in time
	 * 
	 * @return a boolean value indicating whether the event occurs during the
	 *               specified TimePoint
	 */
	public boolean isOccuring(Event anEvent, TimePoint aTimePoint);

	/**
	 * Determines if a given event occurs within a particular date range for
	 * this schedule.
	 * 
	 * @param anEvent
	 *                    the Event whose schedule is being queried
	 * @param aDateRange
	 *                    the time period in which to check schedule information
	 * 
	 * @return a boolean value indicating whether the event occurs within the
	 *               specified DateRange
	 */
	public boolean isOccuring(Event anEvent, DateRange aDateRange);

	/**
	 * Returns a {@link java.util.List List}of
	 * {@link org.chronicj.TimePoint TimePoint}s on which the supplied event
	 * occurs given the specified {@link DateRange}.
	 * 
	 * @param anEvent
	 *                    the Event whose schedule is being queried
	 * @param aDateRange
	 *                    the time period in which to check schedule information
	 * 
	 * @return a boolean value indicating whether the event occurs within the
	 *               specified DateRange
	 */
	public List dates(Event anEvent, DateRange during);

	/**
	 * Returns a {@link TimePoint}indicating the next occurence of the
	 * supplied {@link Event}on or after the given point in time.
	 * 
	 * @param anEvent
	 *                    the Event whose schedule is being queried
	 * @param aTimePoint
	 *                    point in time after which to check the Event's next
	 *                    occurrence.
	 * 
	 * @return a boolean value indicating whether the event occurs within the
	 *               specified DateRange
	 */
	public TimePoint nextOccurence(Event anEvent, TimePoint aTimePoint);

	/**
	 * Adds an {@link Event Event}whose occurance is determined by the given
	 * {@link TemporalExpression}.
	 * 
	 * @param anEvent
	 *                    the Event to add
	 * @param aTemporalExpression
	 *                    the date/time(s) when this Event occurs.
	 */
	public void add(Event anEvent, TemporalExpression aTemporalExpression);
}
