/*
 * ==================================================================== The
 * Apache Software License, Version 1.1
 *
 * Copyright (c) 2003 Digital Clash LLC. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 1.
 * Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer. 2. Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution. 3. The end-user documentation
 * included with the redistribution, if any, must include the following
 * acknowledgment: "This product includes software developed by the ChronicJ
 * team (http://www.chronicj.org/)." Alternately, this acknowledgment may
 * appear in the software itself, if and wherever such third-party
 * acknowledgments normally appear. 4. The names "ChronicJ" and "Digital Clash"
 * not be used to endorse or promote products derived from this software
 * without prior written permission. For written permission, please contact
 * info@digitalclash.com. 5. Products derived from this software may not be
 * called "ChronicJ", "Digital Clash", nor may "ChronicJ" or "Digital Clash"
 * appear in their name, without prior written permission of Digital Clash LLC.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * APACHE SOFTWARE FOUNDATION OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ==================================================================== This
 * product includes software developed by the by the Apache Software Foundation
 * (http://www.apache.org/).
 * ====================================================================
 */
package org.chronicj.impl;

import org.chronicj.TemporalExpression;
import org.chronicj.TimePoint;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;


/**
 * <p>
 * Abstract base class for allowing temporal expressions to be composed of 1..
 * other temporal expressions which can then be used in set expressions.
 * </p>
 *
 * <p>
 * This class is based directly on patterns described in a paper by Martin
 * Fowler which can be found <a
 * href="http://martinfowler.com/apsupp/recurring.pdf">here</a>.
 * </p>
 *
 * @author <a href="mlipper@US-ABP.com">Matthew Lipper</a>
 *
 * @see <a href="http://martinfowler.com/apsupp/recurring.pdf">Recurring
 *      Events for Calendars</a> by Martin Fowler
 */
abstract class CollectionTE implements TemporalExpression {
    private final List expressions;

    /**
     * Creates an instance of a CollectionTE
     */
    CollectionTE() {
        super();
        expressions = new ArrayList();
    }

    /*
     * @see org.chronicj.TemporalExpression#includes(org.chronicj.TimePoint)
     */
    public abstract boolean includes(TimePoint aTimePoint);

    /**
     * Getter for expressions.
     *
     * @return a List of temporal expression which compose this expression
     */
    public List getExpressions() {
        return expressions;
    }

    /**
     * Adds a TemporalExpression to be processed as part of the composition of
     * this TemporalExpression.
     *
     * @param anExpression
     *            the TemporalExpression to add
     *
     * @return a CollectionTE reference to the current instance (<code>this</code>),
     *         enabling easier syntax for building expressions
     */
    public CollectionTE add(TemporalExpression anExpression) {
        expressions.add(anExpression);

        return this;
    }

    /**
     * Getter for querying whether this (abstract) expression has been
     * configured with any actual expressions.
     *
     * @return true if 1 or more TemporalExpression has been added to this
     *         CollectionTE
     */
    public boolean containsExpressions() {
        //expressions is final and initialized in ctor, it should never be
        // null
        return expressions.size() > 0;
    }

    /*
     * (non-Javadoc)
     *
     * @see java.lang.Object#toString()
     */
    public String toString() {
        StringBuffer result = new StringBuffer();
        int count = 0;

        for (Iterator iter = expressions.iterator(); iter.hasNext(); count++) {
            if (count == 0) {
                result.append(iter.next());
            } else {
                result.append(" " + iter.next());
            }
        }

        return result.toString();
    }
}
