/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2003 Digital Clash LLC.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        ChronicJ team (http://www.chronicj.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "ChronicJ" and "Digital Clash" not be used to endorse or
 *                promote products derived from this software without prior written
 *                permission. For written permission, please contact
 *                info@digitalclash.com.
 *
 * 5. Products derived from this software may not be called "ChronicJ",
 *    "Digital Clash", nor may "ChronicJ" or "Digital Clash" appear in
 *                their name, without prior written permission of Digital Clash LLC.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *    This product includes software developed by the by the
 *           Apache Software Foundation  (http://www.apache.org/).
 * ====================================================================
 */
package org.chronicj.impl;

import org.chronicj.TimePoint;

import java.util.Calendar;


/**
 * <p>
 * Implements support for temporal expressions of the form: "last Friday of the
 * month" or "first Tuesday of the month". Expressions syntax is specified in
 * the constructor call ({@link #DayInMonthTE(int, int)}).
 * </p>
 *
 * <p>
 * This class is based directly on patterns described in a paper by  Martin
 * Fowler which can be found  <a
 * href="http://martinfowler.com/apsupp/recurring.pdf">here</a>.
 * </p>
 *
 * @author <a href="mlipper@US-ABP.com">Matthew Lipper</a>
 *
 * @see <a href="http://martinfowler.com/apsupp/recurring.pdf">Recurring Events
 *      for Calendars</a> by Martin Fowler
 */
public class DayInMonthTE extends AbstractTemporalExpression {
    private boolean ignoreWeek;
    private int count;
    private int dayIndex;

    /**
     * Creates a temporal expression using day of the week 1 to 7 and day of week
     * in month which may be either positive or negative indicating whether
     * count is from beginning or end of the month, respectively.  For example,
     * if the first day of the week is set to Sunday  (set in {@link
     * java.util.GregorianCalendar}):<br>
     *
     * <p>
     * <code>TemporalExpression anExpression = new DayInMonthTE(3,2); //example 1</code><br>
     * <code>TemporalExpression anotherExpression = new DayInMonthTE(6,-1);
     * //example 2 </code>
     * </p>
     * Example 1 creates the expression "the second Tuesday" of the month.
     * Example 2 creates the expression "the last Friday" of the month.
     *
     * @param dayIndex the day of the week numbered from 1 to 7
     * @param count the week number from the beginning of the month if positive
     *        or the the week number from the end of the  month if negative.
     */
    public DayInMonthTE(int dayIndex, int count) {
        super();
        this.dayIndex = dayIndex;
        this.count = count;
        ignoreWeek = false;
    }

    /**
     * Creates a temporal expression using day of the week 1 to 7. For
     * example:<br>
     *
     * <p>
     * <code>TemporalExpression sunday = new DayInMonthTE(1); //Sunday</code><br>
     * <code>TemporalExpression saturday = new DayInMonthTE(7);
     * //Saturday</code><code>TemporalExpression thursday = new DayInMonthTE(5);
     * //Thursday</code>
     * </p>
     *
     * @param dayIndex the day of the week numbered from 1 to 7
     */
    public DayInMonthTE(int dayIndex) {
        this(dayIndex, 1);
        ignoreWeek = true;
    }

    /**
     * Does the value of the supplied argument match this expression?  If this
     * instance was created with {@link DayInMonthTE#DayInMonthTE(int, int)}
     * constructor: <br/
     * > true = day of week matches and week of month matches, e.g. 3rd Tuesday
     * of  the month.  If this instance was created with {@link
     * DayInMonthTE#DayInMonthTE(int)} constructor: <br/
     * > true = day of week matches, e.g. Friday.
     *
     * @param aTimePoint the TimePoint against which to check
     *
     * @return boolean true if the given TimePoint occurs during the time this
     *         expression describes.
     *
     * @see org.chronicj.TemporalExpression
     */
    public boolean includes(TimePoint aTimePoint) {
        if (ignoreWeek) {
            return dayMatches(aTimePoint);
        }

        return dayMatches(aTimePoint) && weekMatches(aTimePoint);
    }

    private boolean dayMatches(TimePoint aTimePoint) {
        return aTimePoint.getDayOfWeek() == dayIndex;
    }

    private int daysLeftInMonth(TimePoint aTimePoint) {
        int actualMaximum = aTimePoint.getCalendar().getActualMaximum(Calendar.DAY_OF_MONTH);

        return actualMaximum - aTimePoint.getDayOfMonth();
    }

    private boolean weekFromEndMatches(TimePoint aTimePoint) {
        int daysFromMonthEnd = daysLeftInMonth(aTimePoint) + 1;

        return weekInMonth(daysFromMonthEnd) == Math.abs(count);
    }

    private boolean weekFromStartMatches(TimePoint aTimePoint) {
        return this.weekInMonth(aTimePoint.getDayOfMonth()) == count;
    }

    private int weekInMonth(int dayNumber) {
        return ((dayNumber - 1) / 7) + 1;
    }

    private boolean weekMatches(TimePoint aTimePoint) {
        if (count > 0) {
            return weekFromStartMatches(aTimePoint);
        } else {
            return weekFromEndMatches(aTimePoint);
        }
    }
}
