/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2003 Digital Clash LLC.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        ChronicJ team (http://www.chronicj.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "ChronicJ" and "Digital Clash" not be used to endorse or
 *                promote products derived from this software without prior written
 *                permission. For written permission, please contact
 *                info@digitalclash.com.
 *
 * 5. Products derived from this software may not be called "ChronicJ",
 *    "Digital Clash", nor may "ChronicJ" or "Digital Clash" appear in
 *                their name, without prior written permission of Digital Clash LLC.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *    This product includes software developed by the by the
 *           Apache Software Foundation  (http://www.apache.org/).
 * ====================================================================
 */
package org.chronicj.impl;

import org.chronicj.DateRange;
import org.chronicj.Event;
import org.chronicj.ScheduleElement;
import org.chronicj.TemporalExpression;
import org.chronicj.TimePoint;


/**
 * Default implementation of a {@link org.chronicj.ScheduleElement
 * ScheduleElement} that  can be used to compose a {@link
 * org.chronicj.Schedule Schedule}.
 *
 * @author <a href="mlipper@US-ABP.com">Matthew Lipper</a>
 *
 * @see org.chronicj.ScheduleElement
 */
public class DefaultScheduleElement implements ScheduleElement {
    private Event event;
    private TemporalExpression expression;

    /**
     * Create a default {@link org.chronicj.ScheduleElement ScheduleElement}
     * using the supplied  {@link org.chronicj.Event Event} and {@link
     * org.chronicj.TemporalExpression TemporalExpression}.
     *
         * @throws NullPointerException if either argument is null
     *
     * @param anEvent the event to be scheduled
     * @param anExpression the TemporalExpression that defines when the event occurs
     */
    public DefaultScheduleElement(Event anEvent, TemporalExpression anExpression) {
        super();

        if (null == anEvent) {
            throw new NullPointerException("Event argument cannot be null.");
        }

        event = anEvent;

        if (null == anExpression) {
            throw new NullPointerException(
                "TemporalExpression argument cannot be null.");
        }

        expression = anExpression;
    }

    /**
     * Default implementation of {@link org.chronicj.ScheduleElement#isOccuring(org.chronicj.Event,org.chronicj.TimePoint)}
     * that will return true if the supplied {@link org.chronicj.Event Event}
     * takes place during the supplied {@link org.chronicj.TimePoint TimePoint}.
     *
     * @param anEvent the Event whose schedule is being queried
     * @param aTimePoint an instant in time
     *
     * @return a boolean value indicating whether the event is occuring at the
     *         specified TimePoint
     *
     * @throws NullPointerException if either argument is null
     *
     * @see org.chronicj.ScheduleElement#isOccuring(org.chronicj.Event,org.chronicj.TimePoint)
     */
    public boolean isOccuring(Event anEvent, TimePoint aTimePoint) {
        if (null == anEvent) {
            throw new NullPointerException("Event argument cannot be null.");
        }

        if (null == aTimePoint) {
            throw new NullPointerException("TimePoint argument cannot be null.");
        }

        if (event.equals(anEvent)) {
            return expression.includes(aTimePoint);
        }

        return false;
    }

    /**
     * Default implementation of {@link org.chronicj.ScheduleElement#isOccuring(org.chronicj.Event,org.chronicj.DateRangeImpl)}
     * that will return the logical AND of calling
     * {@link DefaultScheduleElement#isOccuring(org.chronicj.Event,org.chronicj.TimePoint)}
     * for the start and end dates of the supplied DateRange.
     *
     * @param anEvent the Event whose schedule is being queried
     * @param aDateRange the time period in which to check
     *
     * @return a boolean value indicating whether the event occurs within the
     *         specified DateRange
     *
     * @throws NullPointerException if either argument is null
     * @throws IllegalArgumentException if the DateRange argument is empty
     *
     * @see org.chronicj.ScheduleElement#isOccuring(org.chronicj.Event,
     *      org.chronicj.DateRangeImpl)
     */
    public boolean isOccuring(Event anEvent, DateRange aDateRange) {
        if (null == anEvent) {
            throw new NullPointerException("Event argument cannot be null.");
        }

        if (null == aDateRange) {
            throw new NullPointerException("DateRange argument cannot be null.");
        }

        if (aDateRange.isEmpty()) {
            throw new IllegalArgumentException(
                "DateRange argument cannot be empty.");
        }

        if (event.equals(anEvent)) {
            return expression.includes(aDateRange.start()) &&
            expression.includes(aDateRange.end());
        }

        return false;
    }
}
