/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2003 Digital Clash LLC.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        ChronicJ team (http://www.chronicj.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "ChronicJ" and "Digital Clash" not be used to endorse or
 *                promote products derived from this software without prior written
 *                permission. For written permission, please contact
 *                info@digitalclash.com.
 *
 * 5. Products derived from this software may not be called "ChronicJ",
 *    "Digital Clash", nor may "ChronicJ" or "Digital Clash" appear in
 *                their name, without prior written permission of Digital Clash LLC.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *    This product includes software developed by the by the
 *           Apache Software Foundation  (http://www.apache.org/).
 * ====================================================================
 */
package org.chronicj.impl;

import org.chronicj.DateRange;
import org.chronicj.TimePoint;

import java.util.Calendar;
import java.util.Date;


/**
 * <p>
 * Implements support for temporal expressions that describe a recurring span
 * of time each day or across multiple days. In pseudocode, this looks like:
 * "starting at 11:34am and ending at 3:50pm" or "From 11pm to 2am".
 * Expression  syntax is specified  using constructors.
 * </p>
 *
 * <p>
 * This class is based directly on patterns described in a paper by  Martin
 * Fowler which can be found  <a
 * href="http://martinfowler.com/apsupp/recurring.pdf">here</a>.
 * </p>
 *
 * @author <a href="mlipper@US-ABP.com">Matthew Lipper</a>
 *
 * @see <a href="http://martinfowler.com/apsupp/recurring.pdf">Recurring Events
 *      for Calendars</a> by Martin Fowler
 */
public class RangeEachDayTE extends AbstractTemporalExpression {
    private static final int CURRENT_DAY = 1;
    private static final int NEXT_DAY = 2;
    private final Calendar calendar;
    private final DateRange range;
    private final boolean spansMidnight;

    /**
     * Creates a new RangeEachDayTE object. If startHour less than or equal to
     * the endHour, then end hour is assumed to belong to the next day. As such,
     * this class can match a maximim range of 24 hours.
     *
     * @param startHour int 0 to 23
     * @param startMinute int 0 to 59
     * @param endHour int 0 to 23
     * @param endMinute int 0 to 59
     */
    public RangeEachDayTE(int startHour, int startMinute, int endHour,
        int endMinute) {
        calendar = Calendar.getInstance();
        resetCalendar(startHour, startMinute, CURRENT_DAY);

        Date startTime = calendar.getTime();
        System.out.println(DateUtil.format(startTime));

        if (spansMidnight(startHour, endHour)) {
            //Treat it as the next day
            resetCalendar(endHour, endMinute, NEXT_DAY);
            spansMidnight = true;
        } else {
            //Same day
            resetCalendar(endHour, endMinute, CURRENT_DAY);
            spansMidnight = false;
        }

        Date endTime = calendar.getTime();

        System.out.println(DateUtil.format(endTime));

        range = new DateRange(startTime, endTime);

        System.out.println(range);
    }

    /*
     * @see org.chronicj.TemporalExpression#includes(org.chronicj.TimePoint)
     */
    public boolean includes(TimePoint aTimePoint) {
        if (spansMidnight && (aTimePoint.getHourOfDay() < 12)) {
            //Treat it as the next day
            resetCalendar(aTimePoint.getHourOfDay(), aTimePoint.getMinute(),
                NEXT_DAY);
        } else {
            //Same day
            resetCalendar(aTimePoint.getHourOfDay(), aTimePoint.getMinute(),
                CURRENT_DAY);
        }

        return range.includes(new TimePoint(calendar.getTime()));
    }

    /* (non-Javadoc)
     * @see java.lang.Object#toString()
     */
    public String toString() {
        return range.toString();
    }

    private void resetCalendar(int hourOfDay, int minute, int dayOfMonth) {
        calendar.clear();
        calendar.set(Calendar.HOUR_OF_DAY, hourOfDay);
        calendar.set(Calendar.MINUTE, minute);
        calendar.set(Calendar.DAY_OF_MONTH, dayOfMonth);
    }

    private boolean spansMidnight(int startHour, int endHour) {
        if (endHour <= startHour) {
            return true;
        }

        return false;
    }
}
