/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2003 Digital Clash LLC.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        ChronicJ team (http://www.chronicj.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "ChronicJ" and "Digital Clash" not be used to endorse or
 *                promote products derived from this software without prior written
 *                permission. For written permission, please contact
 *                info@digitalclash.com.
 *
 * 5. Products derived from this software may not be called "ChronicJ",
 *    "Digital Clash", nor may "ChronicJ" or "Digital Clash" appear in
 *                their name, without prior written permission of Digital Clash LLC.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *    This product includes software developed by the by the
 *           Apache Software Foundation  (http://www.apache.org/).
 * ====================================================================
 */
package org.chronicj.impl;

import org.chronicj.TimePoint;


/**
 * <p>
 * Implements support for temporal expressions that describe a recurring range
 * of the year. In pseudocode, this is of the form: "starting May 5th and
 * ending  August 28th" or "From November to December". Expression syntax is
 * specified  using constructors.
 * </p>
 *
 * <p>
 * This class is based directly on patterns described in a paper by  Martin
 * Fowler which can be found  <a
 * href="http://martinfowler.com/apsupp/recurring.pdf">here</a>.
 * </p>
 *
 * @author <a href="mlipper@US-ABP.com">Matthew Lipper</a>
 *
 * @see <a href="http://martinfowler.com/apsupp/recurring.pdf">Recurring Events
 *      for Calendars</a> by Martin Fowler
 */
public class RangeEachYearTE extends AbstractTemporalExpression {
    private int endDay;
    private int endMonth;
    private int startDay;
    private int startMonth;

    public RangeEachYearTE(int startMonth, int startDay, int endMonth,
        int endDay) {
        this.startMonth = startMonth;
        this.endMonth = endMonth;
        this.startDay = startDay;
        this.endDay = endDay;
    }

    public RangeEachYearTE(int startMonth, int endMonth) {
        this.startMonth = startMonth;
        this.endMonth = endMonth;
        this.startDay = 0;
        this.endDay = 0;
    }

    public RangeEachYearTE(int month) {
        this.startMonth = month;
        this.endMonth = month;
        this.startDay = 0;
        this.endDay = 0;
    }

    /*
     * @see org.chronicj.TemporalExpression#includes(org.chronicj.TimePoint)
     */
    public boolean includes(TimePoint aTimePoint) {
        return monthsInclude(aTimePoint) || startMonthIncludes(aTimePoint) ||
        endMonthIncludes(aTimePoint);
    }

    private boolean endMonthIncludes(TimePoint aTimePoint) {
        if (aTimePoint.getMonth() != endMonth) {
            return false;
        }

        if (endDay == 0) {
            return true;
        }

        return (aTimePoint.getDayOfMonth() <= endDay);
    }

    private boolean monthsInclude(TimePoint aTimePoint) {
        int month = aTimePoint.getMonth();

        return ((month > startMonth) && (month < endMonth));
    }

    private boolean startMonthIncludes(TimePoint aTimePoint) {
        if (aTimePoint.getMonth() != startMonth) {
            return false;
        }

        if (startDay == 0) {
            return true;
        }

        return (aTimePoint.getDayOfMonth() >= startDay);
    }
}
